/** Single Point of contact for crunching */
export default class PivotData {
    /** Lowest Timestamp of the data supplied */
    startDateTs: number;
    /** Highest Timestamp of the data supplied */
    endDateTs: number;
    /** Selected Row Cols and aggregator chosen in array of arrays */
    selectedRowsCols: string[][];
    /**
     * Calculate Values flag- It keeps the track of last time data was crunched.
     * So in case when only config options needs to changed we wont recalculate data
     */
    calculateValues: boolean;
    /** All Keys of flat data */
    private allAttrs;
    /** Flat Json RAW data - Array<{timestamp:number,Value:number, ...}> */
    private pivotData;
    /** When some kind of filter is applied that filtered data in maintained in this array - same format as pivotData */
    private filteredDataSet;
    /** Array of attributes applied to data - X dimension */
    private rowAttrs;
    /** Array of attributes applied to data - Y dimension */
    private colAttrs;
    /**
     * Map of attribute to inclusion array
     * eg State -> ['Delhi'] etc
     *
     * Filter applied on this attributes
     */
    inclusions: Map<string, string[]>;
    /**
     * Some attributes from allAttrs is not supposed to used for grouping etc.
     * So these attributes are hidden from use perspective
     */
    private hiddenAttrs;
    /**
     * Map of Unit to metric (Special Case)
     * Raw data has a field - Metric which is usually a paramter with possible unit
     * Its a mapping to Metric to Unit
     */
    unitMap: Map<string, string>;
    /** Constructor */
    constructor(pivotData: any[], filters?: any, unitMap?: Map<string, string>);
    /**
     * Filter Data based on inclusions previously set
     * It uses crossfilter which is multi-dimensional filter library.
     * @param {string=} [sortBy=id] Sort the filtered Data on some property by default filtered on id
     * @returns {any[]} filtered dataset sorted on a paramter
     */
    filterDataSet(sortBy?: string): object[];
    /**
     * Group DataSet using Datalib based on certain attributes
     * @param {boolean} [justGroup=true] whether use any custom function on output or Just group the data
     * @param {string=} timeLineAttr If any TimeLineAttr present
     * @param {string=} operation operation to perform eg. - sum,mean,max,min etc
     * @param {string=} attribute attribute to perform aggregation on (Should have number values)
     * @returns {any[]} grouped Data based on above func parameter
     */
    groupDataSet(justGroup?: boolean, timeLineAttr?: string, operation?: string, attribute?: string): object[];
    /**
     * It perform a nesting operation on Raw data.
     * Given the array of keys, it nests data in the order of keys.
     * Where each key is sorted either ascending or specific case
     *
     * This data can be ingested by functions
     *
     * @param {string[]} attrs keys to nest data on
     * @param {any[]} [data=this.filteredDataSet] base data to perform nesting on, i.e. default is filtered data
     * @returns {any[]} nested keys in order provided and also each level is sorted
     */
    private getKeysAttrsV3(attrs, data?);
    /**
     * Reduce the d3 Nest in simpler form.
     * When getKeysAttrsV3 completes the nest the format of data is not ingestible.
     * We call this function iteratively to change format of data
     * @param {any[]} array the input nested array
     * @param {any[]} [element=[]] result of each iteration
     * @returns {any[]} ingestible formated data
     */
    private retrieveValue(array, element?);
    /**
     * Sets filter on data on given attribute
     * Inclusion only keeps track of included selection in case nothing a selected a empty array is added
     *
     * @param {string} attr Key on which filter is applied
     * @param {string[]} selected Array of items selected for above key
     */
    setFilterOnAttr(attr: string, selected: string[]): void;
    /**
     * Getter for row keys
     * @returns {any[]} Nested keys or RowAttrs - X dimension
     */
    getRowKeys(): any[];
    /**
     * Getter for Col keys
     * @returns {any[]} Nested keys or ColAttrs - Y dimension
     */
    getColKeys(): any[];
    /**
     * Initialize allAttrs variable with all possible keys in data
     */
    private setKeys();
    /**
     * Getter for allAttrs
     * Also it hides the hidden Attributes mentioned above
     * @returns {string[]} array of all attributes on data
     */
    readonly allAttr: string[];
    /**
     * Setter for rowAttrs - X dimension
     * It also calculated whether this change would make me recalculate data
     * @param {string[]} rowAttrs Array to keys to set on data
     */
    setRowAttrs(rowAttrs: string[]): void;
    /**
     * Setter for colAttrs - Y dimension
     * It also calculated whether this change would make me recalculate data
     * @param {string[]} colAttrs Array to keys to set on data
     */
    setColAttrs(colAttrs: string[]): void;
    /**
     * Checks if we need to recalculate values
     * @param {boolean} initState Current state of current value
     * @param {string[]} newArray New Value array
     * @param {boolean} isRow Whether working on Row Or Col Attr
     * @returns {boolean} refreshValues or not
     */
    private refreshNeeded(initState, newArray, isRow?);
    /**
     * @returns {string[]} Getter Row Attributes
     */
    getRowAttrs(): string[];
    /**
     * @returns {string[]} Getter Col Attributes
     */
    getColAttrs(): string[];
    /**
     * @returns {any[]} Getter Raw data
     */
    getData(): any[];
    /**
     * Get Keys for one attribute in all Raw data
     * @param {string} attr attribute to get
     * @param {boolean} getAll whether to get all keys or only selected
     * @returns {string[]} keys in string array
     */
    getKeysOneAttrV2(attr: string, getAll?: boolean): string[];
    /**
     * Check if any filter is applied on certain attribute
     * @param {string} attr Attribute to check
     * @returns {boolean} whether any filter is applied or not
     */
    isFilterApplied(attr: string): boolean;
    /**
     * Get display name of all possible aggregator on data
     * @return {string[]} string array of all possible aggregator
     */
    getAgg(): string[];
    /**
     * Performs a operation on single array of numbers
     * @param {any[]} array object array to perform operation on
     * @param {string} operation Operation to perform eg. Sum,Minimum etc
     * @param {string} attribute Attribute to pick from array
     * @returns {number| number[]} based on operation
     */
    getAppropriateTotalBareBone(array: any[], operation: string, attribute: string): number | number[];
    /**
     * Creates Widgets of data
     * 1. Groups data on given attr
     * 2. Performs operations and reduce to array
     * @param {string} attrs attributes array to group on
     * @param {{ attr: string; agg: string }[]=} attrsWithAgg it can also work with individual operation on each attribute
     * @returns {Promise<any>} Promise with returned values
     */
    getWidgetsWithData(attrs: string[], attrsWithAgg?: Array<{
        attr: string;
        agg: string;
    }>): Promise<any>;
    /**
     * Create Attribute to aggregate to nest using D3
     * @param {{ attr: string; agg: string }[]} attrs attribute to aggregator array
     * @param {any[]} [dataSet=this.filteredDataSet()] dataset to work on
     * @returns {number} Crunched value - Single Value
     */
    private createAttrToAggNest(attrs, dataSet?);
    /**
     * Reduce nest and rounding off values array to single value
     * Recursively called
     * @param {any[]} inputResult Input from previous function
     * @param {{ attr: string; agg: string }[]} attrs attribute to aggregator array
     * @param {number} index Current iteration depth - Which attribute to round with which aggregator
     * @returns {number} Rounded value for that level
     */
    private reduceNest(inputResult, attrs, index?);
    /**
     * Quick Group map
     * Takes input returns array of arrays
     * @param {any[]} inputArray Input to work on
     * @param {string} param Attribute to perform it on
     * @param {string} operation Operation to perform
     * @returns {any[][]} Array of Arrays [['Haryana',1212],['Delhi',121],['UP',1212].....]
     */
    private quickGroupMap(inputArray, param, operation?);
    /**
     * @returns {boolean} is object valid
     */
    isValid(): boolean;
    /**
     *
     * @param renderer
     * @returns
     */
    getChosenFilters(renderer?: string): object | string;
    /**
     *
     * @param val
     */
    setData(val: any): void;
    /**
     *
     * @param newMap
     */
    setUnitMap(newMap: Map<string, string>): void;
    /**
     *
     * @param strMap
     */
    strMapToObj(strMap: Map<any, any>): object | boolean;
    /**
     *
     * @param object
     */
    objToStrMap(object: any): Map<any, any>;
    /**
     *
     */
    private setStartEndDate();
}
