(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('crossfilter'), require('d3-array'), require('d3-collection'), require('datalib'), require('lodash')) :
	typeof define === 'function' && define.amd ? define(['crossfilter', 'd3-array', 'd3-collection', 'datalib', 'lodash'], factory) :
	(global.zenCrunchCore = factory(global.crossfilter,global.d3Array,global.d3Collection,global.dl,global._));
}(this, (function (crossfilter,d3Array,d3Collection,dl,_) { 'use strict';

crossfilter = crossfilter && crossfilter.hasOwnProperty('default') ? crossfilter['default'] : crossfilter;

/** Single Point of contact for crunching */
var PivotData = /** @class */ (function () {
    /** Constructor */
    function PivotData(pivotData, filters, unitMap) {
        /** Lowest Timestamp of the data supplied */
        this.startDateTs = 0;
        /** Highest Timestamp of the data supplied */
        this.endDateTs = 0;
        // TODO: Replace with interface
        /** Selected Row Cols and aggregator chosen in array of arrays */
        this.selectedRowsCols = [[], [], ['']];
        /**
         * Calculate Values flag- It keeps the track of last time data was crunched.
         * So in case when only config options needs to changed we wont recalculate data
         */
        this.calculateValues = true;
        /** Flat Json RAW data - Array<{timestamp:number,Value:number, ...}> */
        this.pivotData = [];
        /** When some kind of filter is applied that filtered data in maintained in this array - same format as pivotData */
        this.filteredDataSet = [];
        /** Array of attributes applied to data - X dimension */
        this.rowAttrs = [];
        /** Array of attributes applied to data - Y dimension */
        this.colAttrs = [];
        /**
         * Map of attribute to inclusion array
         * eg State -> ['Delhi'] etc
         *
         * Filter applied on this attributes
         */
        this.inclusions = new Map();
        /**
         * Some attributes from allAttrs is not supposed to used for grouping etc.
         * So these attributes are hidden from use perspective
         */
        this.hiddenAttrs = ['id', 'timestamp', 'Value'];
        /**
         * Map of Unit to metric (Special Case)
         * Raw data has a field - Metric which is usually a paramter with possible unit
         * Its a mapping to Metric to Unit
         */
        this.unitMap = new Map();
        if (filters !== undefined) {
            this.inclusions = this.objToStrMap(filters);
        }
        if (unitMap !== undefined) {
            this.unitMap = unitMap;
        }
        this.setData(pivotData);
    }
    /**
     * Filter Data based on inclusions previously set
     * It uses crossfilter which is multi-dimensional filter library.
     * @param {string=} [sortBy=id] Sort the filtered Data on some property by default filtered on id
     * @returns {any[]} filtered dataset sorted on a paramter
     */
    PivotData.prototype.filterDataSet = function (sortBy) {
        if (sortBy === void 0) { sortBy = 'id'; }
        var mainDataSet = crossfilter(this.pivotData);
        var main = mainDataSet.dimension(function (d) { return d[sortBy] || ''; });
        var timeStamp = mainDataSet.dimension(function (d) { return d.timestamp; });
        this.inclusions.forEach(function (value, fields) {
            value = value.map(function (item) { return String(item); });
            if (!fields.includes('Aggregator')) {
                var dimension = mainDataSet.dimension(function (d) { return d[fields] || []; });
                dimension.filterFunction(function (d) { return value.indexOf(String(d)) > -1; });
            }
        });
        return timeStamp.bottom(Infinity);
    };
    /**
     * Group DataSet using Datalib based on certain attributes
     * @param {boolean} [justGroup=true] whether use any custom function on output or Just group the data
     * @param {string=} timeLineAttr If any TimeLineAttr present
     * @param {string=} operation operation to perform eg. - sum,mean,max,min etc
     * @param {string=} attribute attribute to perform aggregation on (Should have number values)
     * @returns {any[]} grouped Data based on above func parameter
     */
    PivotData.prototype.groupDataSet = function (justGroup, timeLineAttr, operation, attribute) {
        if (justGroup === void 0) { justGroup = true; }
        var groupOnMe = _.concat(this.rowAttrs, this.colAttrs);
        if (timeLineAttr !== undefined && timeLineAttr.length > 0) {
            groupOnMe.push(timeLineAttr);
        }
        var data = this.filterDataSet();
        if (justGroup) {
            return dl.groupby(groupOnMe).execute(data);
        }
        else {
            if (attribute === undefined) {
                throw new Error('Group called without custom flag But Attribute not provided');
            }
            if (operation === undefined) {
                throw new Error('Group called without custom flag But Operation not provided');
            }
            if (operation === 'count') {
                return dl.groupby(groupOnMe)
                    .count()
                    .execute(data);
            }
            return dl.groupby(groupOnMe)
                .summarize([{ name: attribute, ops: [operation] }])
                .execute(data);
        }
    };
    /**
     * It perform a nesting operation on Raw data.
     * Given the array of keys, it nests data in the order of keys.
     * Where each key is sorted either ascending or specific case
     *
     * This data can be ingested by functions
     *
     * @param {string[]} attrs keys to nest data on
     * @param {any[]} [data=this.filteredDataSet] base data to perform nesting on, i.e. default is filtered data
     * @returns {any[]} nested keys in order provided and also each level is sorted
     */
    PivotData.prototype.getKeysAttrsV3 = function (attrs, data) {
        if (data === void 0) { data = this.filteredDataSet; }
        var dayArray = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
        var nestedData = d3Collection.nest();
        _.forEach(attrs, function (attr) {
            if (attr === 'Day') {
                nestedData = nestedData.key(function (d) { return d[attr]; }).sortKeys(function (a, b) {
                    var indexA = dayArray.indexOf(a);
                    var indexB = dayArray.indexOf(b);
                    if (indexA === indexB) {
                        return 0;
                    }
                    else if (indexA < indexB) {
                        return -1;
                    }
                    else {
                        return 1;
                    }
                });
            }
            else {
                nestedData = nestedData.key(function (d) { return d[attr]; }).sortKeys(d3Array.ascending);
            }
        });
        var crunchedData = nestedData.rollup(function (d) { return d['Value']; }).entries(data);
        return this.retrieveValue(crunchedData);
    };
    /**
     * Reduce the d3 Nest in simpler form.
     * When getKeysAttrsV3 completes the nest the format of data is not ingestible.
     * We call this function iteratively to change format of data
     * @param {any[]} array the input nested array
     * @param {any[]} [element=[]] result of each iteration
     * @returns {any[]} ingestible formated data
     */
    PivotData.prototype.retrieveValue = function (array, element) {
        var _this = this;
        if (element === void 0) { element = []; }
        var returnMe = [];
        _.forEach(array, function (val) {
            if (val.values !== undefined) {
                element.push(val.key);
                var returnedKeys = _this.retrieveValue(val.values, element);
                returnMe = _.concat(returnMe, returnedKeys);
            }
            else {
                var newEl = _.clone(element);
                newEl.push(val.key);
                returnMe.push(newEl);
            }
        });
        element.pop();
        return returnMe;
    };
    /**
     * Sets filter on data on given attribute
     * Inclusion only keeps track of included selection in case nothing a selected a empty array is added
     *
     * @param {string} attr Key on which filter is applied
     * @param {string[]} selected Array of items selected for above key
     */
    PivotData.prototype.setFilterOnAttr = function (attr, selected) {
        if (selected === undefined) {
            this.inclusions.delete(attr);
        }
        else {
            this.inclusions.set(attr, selected);
        }
        this.filteredDataSet = this.filterDataSet();
        this.calculateValues = true;
    };
    /**
     * Getter for row keys
     * @returns {any[]} Nested keys or RowAttrs - X dimension
     */
    PivotData.prototype.getRowKeys = function () {
        return this.getKeysAttrsV3(this.rowAttrs);
    };
    /**
     * Getter for Col keys
     * @returns {any[]} Nested keys or ColAttrs - Y dimension
     */
    PivotData.prototype.getColKeys = function () {
        return this.getKeysAttrsV3(this.colAttrs);
    };
    /**
     * Initialize allAttrs variable with all possible keys in data
     */
    PivotData.prototype.setKeys = function () {
        if (this.pivotData.length > 0) {
            var element = this.pivotData[0];
            this.allAttrs = _.keys(element);
        }
    };
    Object.defineProperty(PivotData.prototype, "allAttr", {
        /**
         * Getter for allAttrs
         * Also it hides the hidden Attributes mentioned above
         * @returns {string[]} array of all attributes on data
         */
        get: function () {
            return _.concat(_.difference(this.allAttrs, this.hiddenAttrs), [
                'Aggregator'
            ]);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Setter for rowAttrs - X dimension
     * It also calculated whether this change would make me recalculate data
     * @param {string[]} rowAttrs Array to keys to set on data
     */
    PivotData.prototype.setRowAttrs = function (rowAttrs) {
        this.rowAttrs = rowAttrs;
        this.calculateValues = this.refreshNeeded(this.calculateValues, rowAttrs, true);
    };
    /**
     * Setter for colAttrs - Y dimension
     * It also calculated whether this change would make me recalculate data
     * @param {string[]} colAttrs Array to keys to set on data
     */
    PivotData.prototype.setColAttrs = function (colAttrs) {
        this.colAttrs = colAttrs;
        this.calculateValues = this.refreshNeeded(this.calculateValues, colAttrs, false);
    };
    /**
     * Checks if we need to recalculate values
     * @param {boolean} initState Current state of current value
     * @param {string[]} newArray New Value array
     * @param {boolean} isRow Whether working on Row Or Col Attr
     * @returns {boolean} refreshValues or not
     */
    PivotData.prototype.refreshNeeded = function (initState, newArray, isRow) {
        if (isRow === void 0) { isRow = false; }
        if (initState) {
            return initState;
        }
        else {
            var existingData = isRow
                ? this.selectedRowsCols[0]
                : this.selectedRowsCols[1];
            if (existingData.length !== newArray.length) {
                return true;
            }
            else {
                return _.difference(newArray, existingData).length > 0;
            }
        }
    };
    /**
     * @returns {string[]} Getter Row Attributes
     */
    PivotData.prototype.getRowAttrs = function () {
        return this.rowAttrs;
    };
    /**
     * @returns {string[]} Getter Col Attributes
     */
    PivotData.prototype.getColAttrs = function () {
        return this.colAttrs;
    };
    /**
     * @returns {any[]} Getter Raw data
     */
    PivotData.prototype.getData = function () {
        return this.pivotData;
    };
    /**
     * Get Keys for one attribute in all Raw data
     * @param {string} attr attribute to get
     * @param {boolean} getAll whether to get all keys or only selected
     * @returns {string[]} keys in string array
     */
    PivotData.prototype.getKeysOneAttrV2 = function (attr, getAll) {
        if (getAll === void 0) { getAll = false; }
        if (attr.includes('Aggregator') && !getAll) {
            if (this.inclusions.has('Aggregator')) {
                var returnMe = this.inclusions.get('Aggregator');
                if (returnMe === undefined) {
                    return [];
                }
                return returnMe;
            }
            return this.getAgg();
        }
        else {
            if (this.inclusions.has(attr) && !getAll) {
                var returnMe = this.inclusions.get(attr);
                if (returnMe === undefined) {
                    return [];
                }
                return returnMe;
            }
            return dl.groupby([attr])
                .count()
                .execute(this.pivotData)
                .map(function (item) { return item[attr]; });
        }
    };
    /**
     * Check if any filter is applied on certain attribute
     * @param {string} attr Attribute to check
     * @returns {boolean} whether any filter is applied or not
     */
    PivotData.prototype.isFilterApplied = function (attr) {
        return this.inclusions.has(attr);
    };
    /**
     * Get display name of all possible aggregator on data
     * @return {string[]} string array of all possible aggregator
     */
    PivotData.prototype.getAgg = function () {
        var opDict = {
            Average: 'mean',
            Count: 'count',
            Sum: 'sum',
            Maximum: 'max',
            Minimum: 'min',
            Median: 'median',
            Variance: 'variance',
            '25 percentile': 'q1',
            '75 percentile': 'q3',
            'Standard deviation': 'stdev',
            'IQR (q3-q1)': 'iqr'
        };
        return _.keys(opDict);
    };
    /**
     * Performs a operation on single array of numbers
     * @param {any[]} array object array to perform operation on
     * @param {string} operation Operation to perform eg. Sum,Minimum etc
     * @param {string} attribute Attribute to pick from array
     * @returns {number| number[]} based on operation
     */
    PivotData.prototype.getAppropriateTotalBareBone = function (array, operation, attribute) {
        var format$$1 = dl.format.number(',.2f');
        var opDict = {
            Count: 'count',
            Sum: 'sum',
            Average: 'mean',
            Maximum: 'max',
            Minimum: 'min',
            Median: 'median',
            Variance: 'variance',
            '25 percentile': 'q1',
            '75 percentile': 'q3',
            'Standard deviation': 'stdev',
            'IQR (q3-q1)': 'iqr',
            BoxPlot: 'boxplot'
        };
        var allData = array;
        var result = 0;
        var resultTemp = [];
        switch (opDict[operation]) {
            case 'count':
                result = dl.count(allData, attribute);
                break;
            case 'sum':
                result = dl.sum(allData, attribute);
                break;
            case 'distinct':
                result = dl.count.distinct(allData, attribute);
                break;
            case 'median':
                result = dl.median(allData, attribute);
                break;
            case 'variance':
                result = dl.variance(allData, attribute);
                break;
            case 'stdev':
                result = dl.stdev(allData, attribute);
                break;
            case 'q1':
                resultTemp = dl.quartile(allData, attribute);
                result = resultTemp[0];
                break;
            case 'q3':
                resultTemp = dl.quartile(allData, attribute);
                result = resultTemp[2];
                break;
            case 'iqr':
                resultTemp = dl.quartile(allData, attribute);
                result = resultTemp[2] - resultTemp[0];
                break;
            case 'min':
                result = dl.min(allData, attribute);
                break;
            case 'max':
                result = dl.max(allData, attribute);
                break;
            case 'mean':
                result = dl.mean(allData, attribute);
                break;
            case 'boxplot':
                result = [
                    dl.min(allData, attribute)
                ].concat(dl.quartile(allData, attribute), [
                    dl.max(allData, attribute)
                ]);
                result = result.map(function (val) { return Number(val.toFixed(2)); });
                break;
            default:
                break;
        }
        if (typeof result === 'number') {
            return isNaN(result) ? NaN : Number(result.toFixed(2));
        }
        else {
            return result;
        }
    };
    /**
     * Creates Widgets of data
     * 1. Groups data on given attr
     * 2. Performs operations and reduce to array
     * @param {string} attrs attributes array to group on
     * @param {{ attr: string; agg: string }[]=} attrsWithAgg it can also work with individual operation on each attribute
     * @returns {Promise<any>} Promise with returned values
     */
    PivotData.prototype.getWidgetsWithData = function (attrs, attrsWithAgg) {
        var _this = this;
        return new Promise(function (resolve, reject) {
            if (attrsWithAgg !== undefined && attrsWithAgg.length > 0) {
                _this.selectedRowsCols = [
                    attrsWithAgg.map(function (i) { return i.attr; }),
                    attrsWithAgg.map(function (i) { return i.agg; }),
                    [''],
                    ['']
                ];
                var returnMe_1 = [];
                _.forEach(dl.groupby(['Metric']).execute(_this.filterDataSet()), function (val) {
                    var value = _this.createAttrToAggNest(attrsWithAgg, val['values']);
                    val['valueAgg'] = value.toFixed(2);
                    val['values'] = [];
                    returnMe_1.push(val);
                });
                return resolve(returnMe_1);
            }
            // Creates a array of data items
            // Hack to make sure that some sparkline is created
            // it tries to take date but if dates are too much it takes time
            // Will fail for daily metrics
            var noOfTime = _this.getKeysOneAttrV2('Time').length;
            var noOfDate = _this.getKeysOneAttrV2('Date').length;
            var returnData = [];
            var groupedData = [];
            if (attrs[0] === 'Aggregator') {
                _this.selectedRowsCols = [_this.rowAttrs, _this.colAttrs, [''], ['']];
                groupedData = _this.groupDataSet();
                var aggList_1 = _this.getKeysOneAttrV2('Aggregator');
                _.forEach(groupedData, function (oneGroup) {
                    _.forEach(aggList_1, function (agg) {
                        var newElem = _.cloneDeep(oneGroup);
                        newElem['Aggregator'] = agg;
                        newElem['valueAgg'] = _this.getAppropriateTotalBareBone(newElem['values'], agg, 'Value');
                        newElem['values'] = _this.quickGroupMap(newElem['values'], noOfDate > noOfTime ? 'Date' : 'Time');
                        returnData.push(newElem);
                    });
                });
                return resolve(returnData);
            }
            else {
                var agg_1 = _this.getAgg()[0];
                if (_this.inclusions.get('Aggregator') !== undefined) {
                    var inclusionAgg = _this.inclusions.get('Aggregator');
                    if (inclusionAgg !== undefined) {
                        agg_1 = inclusionAgg[0];
                    }
                }
                _this.selectedRowsCols = [_this.rowAttrs, _this.colAttrs, [''], [agg_1]];
                groupedData = _this.groupDataSet();
                _.forEach(groupedData, function (oneGroup) {
                    oneGroup['Aggregator'] = agg_1;
                    oneGroup['valueAgg'] = _this.getAppropriateTotalBareBone(oneGroup['values'], agg_1, 'Value');
                    oneGroup['values'] = _this.quickGroupMap(oneGroup['values'], noOfDate > noOfTime ? 'Date' : 'Time');
                });
                return resolve(groupedData);
            }
        });
    };
    /**
     * Create Attribute to aggregate to nest using D3
     * @param {{ attr: string; agg: string }[]} attrs attribute to aggregator array
     * @param {any[]} [dataSet=this.filteredDataSet()] dataset to work on
     * @returns {number} Crunched value - Single Value
     */
    PivotData.prototype.createAttrToAggNest = function (attrs, dataSet) {
        if (dataSet === void 0) { dataSet = this.filterDataSet(); }
        var result = d3Collection.nest();
        _.forEach(attrs, function (opToDo) {
            result = result.key(function (d) { return d[opToDo.attr]; });
        });
        var newResult = result.entries(dataSet);
        return this.reduceNest(newResult, attrs, -1);
    };
    /**
     * Reduce nest and rounding off values array to single value
     * Recursively called
     * @param {any[]} inputResult Input from previous function
     * @param {{ attr: string; agg: string }[]} attrs attribute to aggregator array
     * @param {number} index Current iteration depth - Which attribute to round with which aggregator
     * @returns {number} Rounded value for that level
     */
    PivotData.prototype.reduceNest = function (inputResult, attrs, index) {
        var _this = this;
        if (index === void 0) { index = 0; }
        if (inputResult.length > 0) {
            if (_.has(inputResult[0], 'key') && !_.has(inputResult[0], 'value')) {
                index += 1;
                var singleGroupMapped = _.map(inputResult, function (singleGroup) {
                    singleGroup.value = _this.reduceNest(singleGroup.values, attrs, index);
                    delete singleGroup.values;
                    return singleGroup;
                });
                return this.reduceNest(singleGroupMapped, attrs, index);
            }
            else {
                var attToDo = attrs[index];
                var agg = attToDo.agg;
                var crunchedValue = void 0;
                if (_.has(inputResult[0], 'Value')) {
                    crunchedValue = this.getAppropriateTotalBareBone(inputResult, agg, 'Value');
                }
                else {
                    crunchedValue = this.getAppropriateTotalBareBone(inputResult, agg, 'value');
                }
                return crunchedValue;
            }
        }
        return 0;
    };
    /**
     * Quick Group map
     * Takes input returns array of arrays
     * @param {any[]} inputArray Input to work on
     * @param {string} param Attribute to perform it on
     * @param {string} operation Operation to perform
     * @returns {any[][]} Array of Arrays [['Haryana',1212],['Delhi',121],['UP',1212].....]
     */
    PivotData.prototype.quickGroupMap = function (inputArray, param, operation) {
        if (operation === void 0) { operation = 'mean'; }
        return _.map(dl.groupby([param])
            .summarize([{ name: 'Value', ops: [operation] }])
            .execute(inputArray), function (val) {
            return [val[param], val[operation + '_Value']];
        });
    };
    /**
     * @returns {boolean} is object valid
     */
    PivotData.prototype.isValid = function () {
        return this.pivotData.length > 0;
    };
    /**
     *
     * @param renderer
     * @returns
     */
    PivotData.prototype.getChosenFilters = function (renderer) {
        if (renderer === void 0) { renderer = ''; }
        var filters = this.strMapToObj(this.inclusions);
        if (filters === false) {
            return 'Empty Filter present';
        }
        var singleAxisSupport = ['pie', 'boxplot', 'single-value'];
        if (singleAxisSupport.indexOf(renderer) === -1 &&
            this.colAttrs.length === 0) {
            // return 'No Column Chosen';
        }
        var finalOptions = {};
        finalOptions['rows'] = this.selectedRowsCols[0];
        finalOptions['cols'] = this.selectedRowsCols[1];
        finalOptions['filters'] = filters;
        finalOptions['aggregator'] = this.selectedRowsCols[3];
        return finalOptions;
    };
    /**
     *
     * @param val
     */
    PivotData.prototype.setData = function (val) {
        this.pivotData = val;
        this.filteredDataSet = val;
        if (this.unitMap.size === 1) {
            if (this.hiddenAttrs.indexOf('Metric') === -1) {
                this.hiddenAttrs.push('Metric');
            }
        }
        else {
            if (this.hiddenAttrs.indexOf('Metric') !== -1) {
                this.hiddenAttrs.splice(this.hiddenAttrs.indexOf('Metric'), 1);
            }
        }
        this.setKeys();
        this.setStartEndDate();
        this.calculateValues = true;
    };
    /**
     *
     * @param newMap
     */
    PivotData.prototype.setUnitMap = function (newMap) {
        this.unitMap = newMap;
    };
    /**
     *
     * @param strMap
     */
    PivotData.prototype.strMapToObj = function (strMap) {
        var obj = {};
        var nullFound = false;
        strMap.forEach(function (value, key) {
            if (value.length === 0 &&
                typeof value !== 'boolean' &&
                typeof value !== 'string') {
                nullFound = true;
            }
            else {
                obj[key] = value;
            }
        });
        return nullFound ? false : obj;
    };
    /**
     *
     * @param object
     */
    PivotData.prototype.objToStrMap = function (object) {
        var map$$1 = new Map();
        _.forEach(object, function (value, key) {
            if (isNaN(Number(key))) {
                map$$1.set(key, value);
            }
            else {
                map$$1.set(Number(key), value);
            }
        });
        return map$$1;
    };
    /**
     *
     */
    PivotData.prototype.setStartEndDate = function () {
        if (this.pivotData.length === 0) {
            return;
        }
        var mainDataSet = crossfilter(this.pivotData);
        var timeStamp = mainDataSet.dimension(function (d) { return d.timestamp; });
        this.startDateTs = timeStamp.bottom(1)[0]['timestamp'];
        timeStamp.filterAll();
        this.endDateTs = timeStamp.top(1)[0]['timestamp'];
    };
    return PivotData;
}());

return PivotData;

})));
//# sourceMappingURL=zen-crunch-core.js.map
